package utils;

import classUtils.putils.LibraryPathUtils;
import futils.Futil;
import gui.In;

import java.io.ByteArrayOutputStream;
import java.io.File;
import java.io.IOException;
import java.io.PrintStream;
import java.util.Enumeration;
import java.util.Properties;
import java.util.StringTokenizer;

import net.sourceforge.jiu.apps.Strings;
import net.sourceforge.jiu.apps.StringIndexConstants;

/**
 * DocJava, Inc. http://www.docjava.com Programmer: dlyon Date: Mar 29,
 * 2004 Time: 3:33:32 PM
 */
public class SystemUtils {
    public static void main(String[] args) {
        //SystemUtils.testPrintProps();     System.loadLibrary("rxtxSerial");
        String libName = "rxtxSerial";
        System.out.println(libName + " is in path:" + isInPath(libName));
        //Guard the input to System.loadLibrary("rxtxSerial");
        if (isInPath(libName)) {

            System.loadLibrary(libName);
        }

    }

    public static void testJavaCommand() {
        System.out.println("java command=" + getJavaCommand());
    }

    public static void testJarSigner() {
        File f = getSdkCommand("jarsigner");
        System.out.println("f=" + f);
    }

    public static String getTmpDir() {
        return System.getProperty("java.io.tmpdir");
    }

    public static String getLibraryPath() {
        return System.getProperty("java.library.path");
    }

    public static boolean isInPath(String library) {
        String libName = System.mapLibraryName(library);
        System.out.println(libName);
        String paths[] = SystemUtils.getLibraryPaths();
        System.out.println("...Paths...");
        PrintUtils.print(paths);
        String pathsSo[] = StringUtils.getStringsThatEndWith(paths, libName);
        System.out.println("...PathsSo...");
        PrintUtils.print(pathsSo);
        if (pathsSo == null) return false;
        if (paths.length == 0) return false;
        return true;
    }

    public static String[] getLibraryPaths() {
        String s = getLibraryPath();
        StringTokenizer st = new StringTokenizer(s, SystemUtils.getPathSeparator());
        int n = st.countTokens();
        String paths[] = new String[n];
        for (int i = 0; i < n; i++)
            paths[i] = st.nextToken();
        return paths;
    }

    public static String getJavaExtensionsDirectory() {
        return System.getProperty("java.ext.dirs");
    }

    public static String getBootPath() {
        return System.getProperty("sun.boot.library.path");
    }

    /**
     * This works on windows, linux and the mac. It has not been tested on
     * all platforms!
     *
     * @return location of the rmic compiler.
     */
    public static File getRmicPath() {
        // first try the windows location
        return getSdkCommand("rmic");
    }

    /**
     * This works on windows, linux and the mac. It has not been tested on
     * all platforms!
     */
    public static File getJarSignerPath() {
        // first try the windows location
        String command = "jarsigner";
        return getSdkCommand(command);

    }

    /**
     * @return the location of the <code>java</code> command.
     */
    public static File getJavaCommand() {
        return getSdkCommand("java");
    }

    private static File getSdkCommand(String command) {
        String sh = getSdkHome();
        String fs = getDirectorySeparator();
        if (OsUtils.isMacOs()) {
            File f = new File(sh + fs + "Commands" + fs + command);
            if (f.exists()) return f;
        }
        sh = sh +
                fs +
                "bin" +
                fs + command;
        File f = new File(sh);
        if (f.exists()) return f;
        // ok, that did not work..hmm
        // lets try the exe version
        sh = sh + ".exe";
        if (f.exists()) return f;
        //  OK, now I am stuck, lets ask the user:
        System.out.println("sdkHome=" + sh);
        f = Futil.getReadFile(
                "SystemUtils error:please locate " +
                        command);
        System.out.println(
                "please add:" +
                        f +
                        "to the " +
                        command +
                        " in SystemUtils");
        return f;
    }

    public static String getSdkHome() {
        String jh = getJavaHome();
        File f = new File(jh);
        return f.getParent();
    }

    public static String getJavaHome() {
        Properties prop = System.getProperties();
        return prop.getProperty("java.home");
    }

    /**
     * @return the location of the program that is running. Typically
     *         something like:<br><code> /User/lyon/current/j4p</code>.
     */
    public static String getUserDir() {
        return System.getProperty("user.dir");
    }

    /**
     * @return the character(s) used to delimit directories in path names.
     *         For example:<br> /home/lyon/foo, the file separator is "/".
     */
    public static String getDirectorySeparator() {
        return System.getProperty("file.separator");
    }

    /**
     * @return return the character use to delimit the paths in a list of
     *         paths.
     */
    public static String getPathSeparator() {
        return System.getProperty("path.separator");
    }

    public static void printProps() {
        System.out.println(getProps());
    }

    public static String getProps() {
        Properties prop = System.getProperties();
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        PrintStream ps = new PrintStream(baos);
        prop.list(ps);
        try {
            baos.close();
        } catch (IOException e) {
            In.message(e);
        }
        return baos.toString();
    }

    public static void testPrintProps() {
        printProps();
        System.out.println("------------------------");
        System.out.println("os.name=" + OsUtils.getOsName());
        System.out.println("isWindows=" + OsUtils.isWindows());
        System.out.println("free ram=" + free());
        System.out.println("userDir=" + getUserDir());
        System.out.println("userHome=" + getUserHome());
        System.out.println("path sep = " + getPathSeparator());
        System.out.println("file sep = " + getDirectorySeparator());
        System.out.println("getSdkHome=" + getSdkHome());
        System.out.println("getRmicPath=" + getRmicPath());
        System.out.println("getClassPath=" + SystemUtils.getClassPath());
        System.out.println(
                "java.library.path=" + SystemUtils.getLibraryPath());
        System.out.println(
                "java.library.path=" +
                        SystemUtils.getTmpDir());
        LibraryPathUtils.printLibraryPaths();


    }

    /**
     * @return a string that describes the JVM's free RAM
     */
    public static String free() {
        futils.Exec e = new futils.Exec();
        return e.getPrintRamString();
    }

    public static void runGc() {
        Runtime rt = Runtime.getRuntime();
        rt.gc();
        long mem = rt.freeMemory();
        System.out.println("Free Memory = " + mem);
    }

    public static void props() {
        Properties props = System.getProperties();
        for (Enumeration e =
                props.propertyNames(); e.hasMoreElements();) {
            String key = (String) e.nextElement();
            Print.println(key + " = " + props.get(key));
        }
    }

    public static void printProp(String p) {
        System.out.println(p + ":" + System.getProperty(p));
    }

    public static String getClassPath() {
        return System.getProperty("java.class.path");
    }

    public static void home() {
        printProp("java.home");
    }

    public static void userDir() {
        printProp("user.dir");
    }

    /**
     * Returns the users home directory. For example: <code>C:\Documents
     * and Settings\dlyon</code>
     */
    public static String getUserHome() {
        return System.getProperty("user.home");
    }


    public static String getSysInfo() {
        return (
                System.getProperty("os.arch") +
                        " on " +
                        System.getProperty("os.name"));
    }

    public static void testRunMain() throws IOException {
        OsUtils.print(
                runMain(net.rmi.rjs.fc.FractalsJob_1.class));
    }

    /**
     * This method runs the main in a class, c, in a different JVM. It uses
     * the classpath of the current run-time environments
     */
    public static String[] runMain(Class classToRun) throws IOException {
        final String absolutePath = getJavaCommand().getAbsolutePath();
        String className = classToRun.getName();
        String classPath = SystemUtils.getClassPath();
        System.out.println("executing class name..." + className);
        return OsUtils.executeCommand(
                absolutePath +
                        " -cp " + classPath + " " + className);
    }

    public static Object[] getNullArray() {
        return null;
    }


    public static String getMemoryInfo(Strings strings) {
        Runtime runtime = Runtime.getRuntime();
        long freeMemory = runtime.freeMemory();
        long totalMemory = runtime.totalMemory();
        long usedMemory = totalMemory - freeMemory;
        return
                strings.get(StringIndexConstants.FREE_MEMORY) + "=" + freeMemory + "\n" +
                        strings.get(StringIndexConstants.USED_MEMORY) + "=" + usedMemory + "\n" +
                        strings.get(StringIndexConstants.TOTAL_MEMORY) + "=" + totalMemory + "\n";
    }

    /**
     * Returns a multiple-line text with information on the Java Virtual Machine,
     * the path settings and the operating system used, regarding the current
     * language by using a {@link net.sourceforge.jiu.apps.Strings} resource.
     *
     * @return system information as String
     * @see System#getProperty(String)
     */
    public static String getSystemInfo(Strings strings) {
        final String[] PROPERTIES =
                {"java.version",
                        "java.vendor",
                        "java.vendor.url",
                        "java.home",
                        "java.vm.specification.version",
                        "java.vm.specification.vendor",
                        "java.vm.specification.name",
                        "java.vm.version",
                        "java.vm.vendor",
                        "java.vm.name",
                        "java.specification.version",
                        "java.specification.vendor",
                        "java.specification.name",
                        "java.class.version",
                        "java.class.path",
                        "os.name",
                        "os.arch",
                        "os.version",
                        "sun.cpu.endian",
                        "sun.cpu.isalist",
                };
        final int[] STRING_INDEX_VALUES =
                {
                        StringIndexConstants.PROPERTY_JAVA_VERSION,
                        StringIndexConstants.PROPERTY_JAVA_VENDOR,
                        StringIndexConstants.PROPERTY_JAVA_VENDOR_URL,
                        StringIndexConstants.PROPERTY_JAVA_HOME,
                        StringIndexConstants.PROPERTY_JAVA_VM_SPECIFICATION_VERSION,
                        StringIndexConstants.PROPERTY_JAVA_VM_SPECIFICATION_VENDOR,
                        StringIndexConstants.PROPERTY_JAVA_VM_SPECIFICATION_NAME,
                        StringIndexConstants.PROPERTY_JAVA_VM_VERSION,
                        StringIndexConstants.PROPERTY_JAVA_VM_VENDOR,
                        StringIndexConstants.PROPERTY_JAVA_VM_NAME,
                        StringIndexConstants.PROPERTY_JAVA_SPECIFICATION_VERSION,
                        StringIndexConstants.PROPERTY_JAVA_SPECIFICATION_VENDOR,
                        StringIndexConstants.PROPERTY_JAVA_SPECIFICATION_NAME,
                        StringIndexConstants.PROPERTY_JAVA_CLASS_VERSION,
                        StringIndexConstants.PROPERTY_JAVA_CLASS_PATH,
                        StringIndexConstants.PROPERTY_OS_NAME,
                        StringIndexConstants.PROPERTY_OS_ARCH,
                        StringIndexConstants.PROPERTY_OS_VERSION,
                        StringIndexConstants.CPU_ENDIANNESS,
                        StringIndexConstants.CPU_ISALIST
                };
        StringBuffer sb = new StringBuffer();
        for (int i = 0; i < PROPERTIES.length; i++) {
            try {
                String result = System.getProperty(PROPERTIES[i]);
                if (result != null) {
                    if (sb.length() > 0) {
                        sb.append("\n");
                    }
                    sb.append(strings.get(STRING_INDEX_VALUES[i]));
                    sb.append('=');
                    sb.append(result);
                }
            }
            catch (Exception e) {
            }
        }
        return sb.toString();
    }
}

/*
java.runtime.name=Java(TM) 2 Runtime Environment, Stand...
sun.boot.library.path=C:\j2sdk1.4.1_02\jre\bin
java.vm.version=1.4.1_02-b06
java.vm.vendor=Sun Microsystems Inc.
java.vendor.url=http://java.sun.com/
path.separator=;
java.vm.name=Java HotSpot(TM) Client VM
file.encoding.pkg=sun.io
idea.launcher.port=7556
user.country=US
sun.os.patch.level=Service Pack 1
java.vm.specification.name=Java Virtual Machine Specification
user.dir=C:\lyon\j4p
java.runtime.version=1.4.1_02-b06
java.awt.graphicsenv=sun.awt.Win32GraphicsEnvironment
java.endorsed.dirs=C:\j2sdk1.4.1_02\jre\lib\endorsed
os.arch=x86
java.io.tmpdir=C:\DOCUME~1\dlyon\LOCALS~1\Temp\
line.separator=

java.vm.specification.vendor=Sun Microsystems Inc.
user.variant=
os.name=Windows XP
sun.java2d.fontpath=
java.library.path=C:\j2sdk1.4.1_02\bin;.;C:\WINDOWS\Sys...
java.specification.name=Java Platform API Specification
java.class.version=48.0
java.util.prefs.PreferencesFactory=java.util.prefs.WindowsPreferencesFac...
os.version=5.1
user.home=C:\Documents and Settings\dlyon
user.timezone=
java.awt.printerjob=sun.awt.windows.WPrinterJob
file.encoding=windows-1252
java.specification.version=1.4
user.name=dlyon
java.class.path=C:\j2sdk1.4.1_02\jre\lib\charsets.jar...
java.vm.specification.version=1.0
sun.arch.data.model=32
java.home=C:\j2sdk1.4.1_02\jre
java.specification.vendor=Sun Microsystems Inc.
user.language=en
awt.toolkit=sun.awt.windows.WToolkit
java.vm.info=mixed mode
java.version=1.4.1_02
java.ext.dirs=C:\j2sdk1.4.1_02\jre\lib\ext
sun.boot.class.path=C:\j2sdk1.4.1_02\jre\lib\rt.jar;C:\j2...
java.vendor=Sun Microsystems Inc.
file.separator=\
java.vendor.url.bug=http://java.sun.com/cgi-bin/bugreport...
idea.launcher.library=C:\IntelliJ-IDEA-4.0\bin\breakgen.dll
sun.cpu.endian=little
sun.io.unicode.encoding=UnicodeLittle
sun.cpu.isalist=pentium i486 i386
*/
